"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.empty = exports.length = exports.assert = exports.assertMonotonic = exports.assertGlobal = exports.assertLocal = exports.delete = exports.count = exports.countLeft_ = exports.each = exports.shift = exports.shiftCount = exports.maybeShiftCount = exports.insert = exports.has = exports.getc = exports.getv = exports.get = exports.of = exports.Cmp = void 0;
const prelude_js_1 = require("./prelude.js");
const Cmp = __importStar(require("@prelude/cmp"));
exports.Cmp = Cmp;
const blacken_js_1 = __importDefault(require("./blacken.js"));
const delete_js_1 = __importDefault(require("./delete.js"));
const insert_js_1 = __importDefault(require("./insert.js"));
const redden_js_1 = __importDefault(require("./redden.js"));
const shift_js_1 = __importDefault(require("./shift.js"));
/** @returns Red-black tree based on element-to-key mapping function and comparision over keys function. */
const of = (cmp, key) => ({
    cmp,
    key,
    root: prelude_js_1.E
});
exports.of = of;
const get_ = (_, key, cmp) => {
    if (_ == null) {
        return [undefined, -0];
    }
    const cmp_ = cmp(key(_.v));
    switch (cmp_) {
        case Cmp.asc:
            return get_(_.l, key, cmp);
        case Cmp.equal:
            return [_.v, _.n];
        case Cmp.dsc:
            return get_(_.r, key, cmp);
        default:
            throw new TypeError(`Expected cmp result, got ${cmp_}.`);
    }
};
const get = (tree, key) => get_(tree.root, tree.key, b => tree.cmp(key, b));
exports.get = get;
const getv = (tree, key) => (0, exports.get)(tree, key)[0];
exports.getv = getv;
const getc = (tree, key) => (0, exports.get)(tree, key)[1];
exports.getc = getc;
const has = (tree, key) => (0, exports.getc)(tree, key) > 0;
exports.has = has;
/**
 * Inserts provided `value`.
 * If element already exists, `merge` is used to produce final element (defaults to identity function).
 */
const insert = (tree, element, n = 1, merge = _ => _) => {
    tree.root = (0, blacken_js_1.default)((0, insert_js_1.default)(tree.root, tree.key, tree.cmp, element, n, merge));
};
exports.insert = insert;
/**
 * Shifts left-most element (ie. minimum element).
 * @returns `undefined` or shifted element.
 */
// TODO: shiftSame
const maybeShiftCount = (tree) => {
    if (tree.root == null) {
        return [undefined, 0];
    }
    const [x, i, a] = (0, shift_js_1.default)(tree.root);
    tree.root = a;
    // ??
    if (tree.root) {
        tree.root.c = prelude_js_1.B;
    }
    return [x, i];
};
exports.maybeShiftCount = maybeShiftCount;
/**
 * @returns removed left-most element.
 * @throws if tree is empty.
 */
const shiftCount = (tree) => {
    const [x, i] = (0, exports.maybeShiftCount)(tree);
    if (x === undefined) {
        throw new Error('Error while trying to shift an empty rb-tree.');
    }
    return [x, i];
};
exports.shiftCount = shiftCount;
const shift = (tree) => (0, exports.shiftCount)(tree)[0];
exports.shift = shift;
const eachNode = function* (n) {
    if (!n) {
        return;
    }
    yield* eachNode(n.l);
    yield n;
    yield* eachNode(n.r);
};
// TODO: eachCount
const each = function* (tree) {
    for (const _ of eachNode(tree.root)) {
        yield _.v;
    }
};
exports.each = each;
const countLeft_ = (_, key, cmp, inclusive) => {
    if (!_) {
        return 0;
    }
    const cmp_ = cmp(key(_.v));
    switch (cmp_) {
        case Cmp.asc:
            return (0, exports.countLeft_)(_.l, key, cmp, inclusive);
        case Cmp.equal:
            return (_.l?.s ?? 0) + (inclusive ? _.n : 0);
        case Cmp.dsc:
            return _.s - (_.r?.s ?? 0) + (0, exports.countLeft_)(_.r, key, cmp, inclusive);
        default:
            throw new TypeError(`Expected cmp result, got ${cmp_}.`);
    }
};
exports.countLeft_ = countLeft_;
const count = (tree, query) => {
    const n = tree.root?.s ?? 0;
    if (query === undefined) {
        return n;
    }
    const cl = (k, e) => (0, exports.countLeft_)(tree.root, tree.key, b => tree.cmp(k, b), e);
    // How many to remove?
    const a = '$l' in query ? n - cl(query.$l, false) : 0;
    const b = '$le' in query ? n - cl(query.$le, true) : 0;
    const c = '$r' in query ? cl(query.$r, true) : 0;
    const d = '$re' in query ? cl(query.$re, false) : 0;
    const x = Math.max(a, b);
    const y = Math.max(c, d);
    return n - x - y;
};
exports.count = count;
const delete_ = (tree, key, i = 1) => {
    const [v, n, r] = (0, delete_js_1.default)((0, redden_js_1.default)(tree.root), tree.key, b => tree.cmp(key, b), i);
    tree.root = r;
    return [v, n];
};
exports.delete = delete_;
/** Assert local invariant - red nodes don't have red children. */
const assertLocal = (tree) => {
    for (const _ of eachNode(tree.root)) {
        if (_.c === prelude_js_1.R) {
            if ((_.l?.c ?? prelude_js_1.B) !== prelude_js_1.B || (_.r?.c ?? prelude_js_1.B) !== prelude_js_1.B) {
                throw new Error('Local invariant violated.');
            }
        }
    }
};
exports.assertLocal = assertLocal;
/** Assert global invariant - paths from nil leafs up to the root go through same number of black nodes. */
const assertGlobal = (tree) => {
    const visit = function* (n, b) {
        if (!n) {
            yield b;
            return;
        }
        const b_ = b + (n.c === prelude_js_1.B ? 1 : 0);
        yield* visit(n.l, b_);
        yield* visit(n.r, b_);
    };
    let b_ = undefined;
    for (const b of visit(tree.root, 0)) {
        if (b_ === undefined) {
            b_ = b;
        }
        if (b_ !== b) {
            throw new Error('Global invariant violated.');
        }
    }
};
exports.assertGlobal = assertGlobal;
const assertMonotonic = (tree) => {
    let k = undefined;
    for (const _ of (0, exports.each)(tree)) {
        const k_ = tree.key(_);
        if (k === undefined) {
            k = k_;
        }
        if (tree.cmp(k, k_) === Cmp.dsc) {
            throw new Error('Monotonic invariant violated.');
        }
    }
};
exports.assertMonotonic = assertMonotonic;
const assert = (tree) => {
    (0, exports.assertLocal)(tree);
    (0, exports.assertGlobal)(tree);
    (0, exports.assertMonotonic)(tree);
};
exports.assert = assert;
/** @returns number of elements. */
const length = (tree) => tree.root?.s ?? 0;
exports.length = length;
/** @returns `true` if there are no elements, `false` otherwise. */
const empty = (tree) => (0, exports.length)(tree) === 0;
exports.empty = empty;
//# sourceMappingURL=rb-tree.js.map