import { inspect } from 'util';
import * as RbTree from './rb-tree.js';
var Cmp = RbTree.Cmp;
export { Cmp };
export const of = (cmp) => ({
    tree: RbTree.of(cmp, (_) => _[0])
});
export const set = (map, key, value) => RbTree.insert(map.tree, [key, value], 0, _ => _);
export const maybeGet = (map, key) => RbTree.get(map.tree, key)[0]?.[1];
export const get = (map, key) => {
    const _ = maybeGet(map, key);
    if (_ === undefined) {
        throw new Error(`Key ${inspect(key)} not found.`);
    }
    return _;
};
export const has = (map, key) => RbTree.has(map.tree, key);
/**
 * Deletes left most element.
 * @returns left most element.
 * @throws if map is empty.
 */
export const shift = (map) => RbTree.shift(map.tree)[1];
/**
 * Iterates over keys.
 * Modifications inside iteration are allowed.
 * Iteration will yield keys as they were during initial invocation.
 */
export const keys = function* (map) {
    for (const _ of RbTree.each(map.tree)) {
        yield _[0];
    }
};
export const values = function* (map) {
    for (const _ of RbTree.each(map.tree)) {
        yield _[1];
    }
};
export const entries = function* (map) {
    for (const _ of RbTree.each(map.tree)) {
        yield [_[0], _[1]];
    }
};
//# sourceMappingURL=map.js.map